# this code reproduces the application in section 7 of the paper

###############
# IMPORTANT NOTE: the real dataset used in the paper is not public: 
# the demo is ran with an artificial dataset (similar to the one used in the paper)
dat <- read.table(file="fake-data.txt")
###############


library(INLA)
library(spam)
source("utilities.R")


# y: response  (equivalent to log(NO3) in the paper)
y <- dat$y

# x: covariate (equivalent to stream distance in the paper)
x <- dat$x

# season is a factor with 4 levels: 
unique(dat$season)

# give a look at the "fake" data
plot(x,y)
points(x[dat$season=="spring"],y[dat$season=="spring"], col=2, pch=16)
points(x[dat$season=="summer"],y[dat$season=="summer"], col=3, pch=16)
points(x[dat$season=="autumn"],y[dat$season=="autumn"], col=4, pch=16)
points(x[dat$season=="winter"],y[dat$season=="winter"], col=5, pch=16)



# build design matrix X and B-spline matrix B (see paper section 7)
# covariate
x.s1 <- x.s2 <- x.s3 <- x.s4 <- x
x.s1[dat$season!="spring"] <- 0
x.s2[dat$season!="summer"] <- 0
x.s3[dat$season!="autumn"] <- 0
x.s4[dat$season!="winter"] <- 0

# factor "season", below this becomes "s" for simplicity
n <- length(y)
a <- rep(1,n)   # intercept associated to season=="spring"
s2 <- s3 <- s4 <- rep(0,n)
s2[dat$season=="summer"] <- 1
s3[dat$season=="autumn"] <- 1
s4[dat$season=="winter"] <- 1

# define fixed effects design matrix X
X <- cbind(a, s2, s3, s4, x.s1, x.s2, x.s3,x.s4)
XtX <- crossprod(X,X)

# set the B-spline design
nseg <- 27
bdeg <- 3
ord <- 2  

B1 <- bspline(x=x.s1, xl=min(x), xr=max(x), ndx=nseg, bdeg=bdeg)
B2 <- bspline(x=x.s2, xl=min(x), xr=max(x), ndx=nseg, bdeg=bdeg)
B3 <- bspline(x=x.s3, xl=min(x), xr=max(x), ndx=nseg, bdeg=bdeg)
B4 <- bspline(x=x.s4, xl=min(x), xr=max(x), ndx=nseg, bdeg=bdeg)
B <- cbind(B1,B2,B3,B4)
K <- ncol(B)
BtB <- crossprod(B,B)

# define structure matrix R; ONLY works with ord=2, IGMRF of order 2
D0 <- diff(diag(ncol(B1)), diff=ord)
R0 <- crossprod(D0,D0)
R <- diag(4) %x% R0

### set constraint matrices A1, A2, A3, A4, one for each seasonal smooth term
e <- matrix(c(0,0),2,1) 

# spring
A11 <- apply(B1,2,sum)
v <- c(x.s1)
A12 <- rowSums(crossprod(B1, diag(v)))
A1 <- rbind(A11,A12)
# NOTE: in our experience, it is convenient to rescale the constraints, like:
A1 <- rbind(A1[1,]/abs(max(A1[1,])), A1[2,]/abs(max(A1[2,]))) 
# This rescaling works well in all the applications that we investigated so far.
# Without this rescaling, we get numerical problems when computing the density of the full conditional for beta under linear constr 
# (which is needed in computing the acceptance ratio for joint (tau.beta, beta))

# summer
A11 <- apply(B2,2,sum)
v <- c(x.s2)
A12 <- rowSums(crossprod(B2, diag(v)))
A2 <- rbind(A11,A12)
A2 <- rbind(A2[1,]/abs(max(A2[1,])), A2[2,]/abs(max(A2[2,]))) 

# autumn
A11 <- apply(B3,2,sum)
v <- c(x.s3)
A12 <- rowSums(crossprod(B3, diag(v)))
A3 <- rbind(A11,A12)
A3 <- rbind(A3[1,]/abs(max(A3[1,])), A3[2,]/abs(max(A3[2,])))  

# winter
A11 <- apply(B4,2,sum)
v <- c(x.s4)
A12 <- rowSums(crossprod(B4, diag(v)))
A4 <- rbind(A11,A12)
A4 <- rbind(A4[1,]/abs(max(A4[1,])), A4[2,]/abs(max(A4[2,]))) 

# build full constraint matrix A
A <- cbind(A1,A2,A3,A4)


# run INLA with prior Gamma(1,5e-5) on tau.beta_j, j=1,...,4 (the 4 season) 
# may be useful to get reasonable initial values for MCMC
stk <- inla.stack(data=list(y=y), 
                  A=list(1,B1,B2,B3,B4),
                  effects=list(list(a=X[,1], 
                                    summer=X[,2], autumn=X[,3], winter=X[,4],
                                    x.spring=X[,5], x.summer=X[,6], 
                                    x.autumn=X[,7], x.winter=X[,8]), 
                               beta.spring=1:ncol(B1),
                               beta.summer=1:ncol(B2),
                               beta.autumn=1:ncol(B3),
                               beta.winter=1:ncol(B4)))
r <- inla(y ~ -1 + a + summer + autumn + winter +
                         x.spring + x.summer + x.autumn + x.winter +
                          f(beta.spring, model="generic0", Cmatrix=as.matrix(R0), rankdef=2,
                            hyper=list(theta=list(prior="loggamma", param=c(1, 5e-5))),
                            constr=FALSE,
                            extraconstr=list(A=matrix(A1[1:2,],2),e=matrix(c(0,0),2,1)))+
                          f(beta.summer, model="generic0", Cmatrix=as.matrix(R0), rankdef=2,
                            hyper=list(theta=list(prior="loggamma", param=c(1, 5e-5))),
                            constr=FALSE,
                            extraconstr=list(A=matrix(A2[1:2,],2),e=matrix(c(0,0),2,1)))+
                          f(beta.autumn, model="generic0", Cmatrix=as.matrix(R0), rankdef=2,
                            hyper=list(theta=list(prior="loggamma", param=c(1, 5e-5))),
                            constr=FALSE,
                            extraconstr=list(A=matrix(A3[1:2,],2),e=matrix(c(0,0),2,1)))+
                          f(beta.winter, model="generic0", Cmatrix=as.matrix(R0), rankdef=2,
                            hyper=list(theta=list(prior="loggamma", param=c(1, 5e-5))),
                            constr=FALSE,
                            extraconstr=list(A=matrix(A4[1:2,],2),e=matrix(c(0,0),2,1))), 
                        data=inla.stack.data(stk), 
                        control.predictor=list(A=inla.stack.A(stk), compute=TRUE),
                        control.compute=list(dic=TRUE),
                        control.family = list(hyper = list(prec = list(prior = "loggamma", param = c(1, 5e-5)))),
                        verbose=FALSE)  



### run MCMC (the code below follows algorithm 2 in the paper)

# first, initialize the MCMC chains (we nitialize to INLA output)
tau.e <- r$summary.hyperpar['Precision for the Gaussian observations',"mean"]
tau.b.spring <-  r$summary.hyperpar['Precision for beta.spring',"mean"]
tau.b.summer <-  r$summary.hyperpar['Precision for beta.summer',"mean"]
tau.b.autumn <- r$summary.hyperpar['Precision for beta.autumn',"mean"]
tau.b.winter <- r$summary.hyperpar['Precision for beta.winter',"mean"]
tau.b <- c(tau.b.spring, tau.b.summer, tau.b.autumn, tau.b.winter)
tau.g <- .001

beta.spring <- r$summary.random$'beta.spring'$mean #rep(0,ncol(B1))
beta.summer <- r$summary.random$'beta.summer'$mean #rep(0,ncol(B1))
beta.autumn <- r$summary.random$'beta.autumn'$mean #rep(0,ncol(B1))
beta.winter <- r$summary.random$'beta.winter'$mean #rep(0,ncol(B1))
beta <- c(beta.spring, beta.summer, beta.autumn, beta.winter)
gamma <- r$summary.fixed[,"mean"] 


# initialize the PCprior
# 1) for each season, compute the mapping h(tau.beta.xxx), conditional to tau.eps
tau.eps <- tau.e
map1 <- h.tau(x=x.s1, nseg=nseg, bdeg=bdeg, diff.ord=ord, tau.e=tau.eps, eps.diag=1e-5)
map2 <- h.tau(x=x.s2, nseg=nseg, bdeg=bdeg, diff.ord=ord, tau.e=tau.eps, eps.diag=1e-5)
map3 <- h.tau(x=x.s3, nseg=nseg, bdeg=bdeg, diff.ord=ord, tau.e=tau.eps, eps.diag=1e-5)
map4 <- h.tau(x=x.s4, nseg=nseg, bdeg=bdeg, diff.ord=ord, tau.e=tau.eps, eps.diag=1e-5)

# 2) for each season, scale the PCprior according to U and alpha
alpha <-0.01
U1<-U2<-U3<-U4 <- 8
# spring
u1 <-  1/sqrt(map1$tau(U1))
theta1 <- -log(alpha)/u1
# summer
u2 <-  1/sqrt(map2$tau(U2))
theta2 <- -log(alpha)/u2
# autumn
u3 <-  1/sqrt(map3$tau(U3))
theta3 <- -log(alpha)/u3
# winter
u4 <-  1/sqrt(map4$tau(U4))
theta4 <- -log(alpha)/u4


# compute the initial current log.marginals for tau.e and tau.b:
# tau.e
mu.tilde <- c(B%*%beta)
Xty <- crossprod(X, y-mu.tilde)
b.fixed <- tau.e*Xty
Q.fixed <- tau.e*XtX + tau.g*diag(ncol(X))
gamma0 <- sample.fixed(Q.fixed, b.fixed)
gamma.cand <- gamma0$gamma
L.fixed <- gamma0$L
z.fixed <- gamma0$z
log.marginal.tau.e.current <- marginal.prec.errors(y, tau.e, mu.tilde, X, 
                                                   gamma.cand, tau.g, L.fixed, z.fixed)  
# tau.b
mu.tilde <- c(X%*%gamma)
Bty <- crossprod(B, y-mu.tilde)
b.spline <- tau.e*Bty
Q.spline <- tau.e*BtB + diag(tau.b)%x%crossprod(D0,D0)
beta0 <- sample.gmrf.ULC(Q.spline, b.spline, A)
beta.cand <- beta0$beta
mean.beta <- beta0$mean.beta  # NOTE: this is the mean of the GMRF unconstrained (see notes)
L.spline <- beta0$L
log.marginal.tau.b.current <- 
  marginal.prec.splinecoef.allthe4.ULC(y, tau.e, mu.tilde, 
                                       B, beta.cand, mean.beta, Q.spline, L.spline, 
                                       A, tau.b, R0, ord, 
                                       theta.e=c(theta1,theta2,theta3,theta4)) 

# build empty mcmm boxes (we run 20000 iteration, thinning by 20)
burnin <- 1000
n.iter <- 20000
thin <- 20
acceptance.fixed <- acceptance.spline <- rep(0, burnin+n.iter)
samples.beta <- matrix(nrow=length(beta), ncol=0)
samples.gamma <- matrix(nrow=length(gamma), ncol=0)
samples.tau.e <- matrix(, nrow=1, ncol=0)
samples.tau.b <- matrix(, nrow=4, ncol=0)
tau.b.cand <- tau.b

verbose <- TRUE
tun <-1.5

for (i in 1:(burnin+n.iter)) {
  
  ### the mcmc is a two block M-H to accept/reject (gamma, tau.e) and (tau.b, beta) jointly 
  tau.e.cand <- tau.e * ranf(1,tun)
  
  ## M-H for (gamma, tau.e)
  # sample 'fixed effects' gamma given tau.e.cand
  mu.tilde <- c(B1%*%beta.spring+B2%*%beta.summer+B3%*%beta.autumn+B4%*%beta.winter)
  Xty <- crossprod(X, y-mu.tilde)
  b.fixed <- tau.e.cand*Xty
  Q.fixed <- tau.e.cand*XtX + tau.g*diag(ncol(X))
  gamma0 <- sample.fixed(Q.fixed, b.fixed)
  gamma.cand <- gamma0$gamma
  L.fixed <- gamma0$L
  z.fixed <- gamma0$z
  # accept/reject (tau.e.cand, gamma)
  log.marginal.tau.e.new <- marginal.prec.errors(y, tau.e.cand, mu.tilde, X, 
                                                 gamma.cand, tau.g, L.fixed, z.fixed)  
  if (runif(1) < exp(min(0, log.marginal.tau.e.new - log.marginal.tau.e.current))){
    log.marginal.tau.e.current <- log.marginal.tau.e.new
    tau.e <- tau.e.cand
    gamma <- gamma.cand
    finvU1 <- (h.tau.given.scale(map1$v0, tau.e)$tau(U1))   # re-scale the PC-prior according to tau.e  
    theta1 <- -log(alpha)*sqrt(finvU1)
    finvU2 <- (h.tau.given.scale(map2$v0, tau.e)$tau(U2))   # re-scale the PC-prior according to tau.e  
    theta2 <- -log(alpha)*sqrt(finvU2)
    finvU3 <- (h.tau.given.scale(map3$v0, tau.e)$tau(U3))   # re-scale the PC-prior according to tau.e  
    theta3 <- -log(alpha)*sqrt(finvU3)
    finvU4 <- (h.tau.given.scale(map4$v0, tau.e)$tau(U4))   # re-scale the PC-prior according to tau.e  
    theta4 <- -log(alpha)*sqrt(finvU4)
    acceptance.fixed[i] <- 1
  }
  
  ## M-H for (beta, tau.b)
  # sample 'spline coefs' for all the 4 seasons
  
  tau.b.cand[1] <- tau.b[1] * ranf(1,tun)
  tau.b.cand[2] <- tau.b[2] * ranf(1,tun)
  tau.b.cand[3] <- tau.b[3] * ranf(1,tun)
  tau.b.cand[4] <- tau.b[4] * ranf(1,tun)
  
  mu.tilde <- c(X%*%gamma)
  Bty <- crossprod(B, y-mu.tilde)
  b.spline <- tau.e*Bty
  Q.spline <- tau.e*BtB + diag(tau.b.cand)%x%crossprod(D0,D0)
  beta0 <- sample.gmrf.ULC(Q.spline, b.spline, A)
  beta.cand <- beta0$beta
  mean.beta <- beta0$mean.beta  # NOTE: this is the mean of the GMRF unconstrained (see notes)
  L.spline <- beta0$L
  log.marginal.tau.b.new <- 
    marginal.prec.splinecoef.allthe4.ULC(y, tau.e, mu.tilde, 
                                         B, beta.cand, mean.beta, Q.spline, L.spline, 
                                         A, tau.b.cand, R0, ord, 
                                         theta.e=c(theta1,theta2,theta3,theta4))
  
  if (runif(1) < exp(min(0, log.marginal.tau.b.new - log.marginal.tau.b.current))){
    log.marginal.tau.b.current <- log.marginal.tau.b.new
    tau.b <- tau.b.cand
    beta <- beta.cand
    acceptance.spline[i] <- 1
  } 
  
  # verbose
  k <- i/100;  
  if(verbose & (ceiling(k) == floor(k))) {
    accept.rate.fixed <- mean(acceptance.fixed[(i-100+1):i])*100
    accept.rate.spline <- mean(acceptance.spline[(i-100+1):i])*100
    print(paste("n.iter ", i,"  ; accept FIXED=", accept.rate.fixed))
    print(paste("n.iter ", i,"  ; accept SPLINE=", accept.rate.spline, sep=""))
  }
  
  # collect samples
  if (i>burnin & (i %% thin)==0){
    samples.tau.e <- cbind(samples.tau.e, tau.e)
    samples.tau.b <- cbind(samples.tau.b, tau.b)
    samples.beta <- cbind(samples.beta, beta)
    samples.gamma <- cbind(samples.gamma, gamma)
  }
}

#  mcmc diagnostic
par(mfrow=c(1,1))
plot(log(samples.tau.e[1,]), type="l")
plot(log(samples.tau.b[1,]), type="l")
plot(log(samples.tau.b[2,]), type="l")
plot(log(samples.tau.b[3,]), type="l")
plot(log(samples.tau.b[4,]), type="l")
par(mfrow=c(3,3))
for(i in 1:8) plot(samples.gamma[i,], type="l")



# compute the smooth f(x) for each season
ind.spring <- 1:30
ind.summer <- 31:60
ind.autumn <- 61:90
ind.winter <- 91:120
samples.beta.spring <- samples.beta[ind.spring,]
samples.beta.summer <- samples.beta[ind.summer,]
samples.beta.autumn <- samples.beta[ind.autumn,]
samples.beta.winter <- samples.beta[ind.winter,]

xg <- seq(min(x), max(x), length.out=500)
Bg <- bspline(x=xg, xl=min(x), xr=max(x), ndx=nseg, bdeg=bdeg)

n.mcmc <- n.iter/thin

# spring
mcmc <- matrix(c(samples.gamma[5,]) %x% xg, 500, n.mcmc) + Bg%*%samples.beta.spring
mcmc.lin <- matrix(c(samples.gamma[5,]) %x% xg, 500, n.mcmc) 
fit.pcprior.mcmc.spring <- apply(mcmc,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.spring <- apply(mcmc,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.spring <- apply(mcmc,1,quantile,0.975, na.rm=TRUE)
fit.pcprior.mcmc.lin.spring <- apply(mcmc.lin,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.lin.spring <- apply(mcmc.lin,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.lin.spring <- apply(mcmc.lin,1,quantile,0.975, na.rm=TRUE)
partial.fit.out.spring <- apply(X[,-5]%*%samples.gamma[-5,] + 
                                     B[,-ind.spring]%*%samples.beta[-ind.spring,], 1, mean)  #  uguale
partial.res.out.spring <- y - (partial.fit.out.spring)

# summer
mcmc <- matrix(c(samples.gamma[6,]) %x% xg, 500, n.mcmc) + Bg%*%samples.beta.summer
mcmc.lin <- matrix(c(samples.gamma[6,]) %x% xg, 500, n.mcmc)
fit.pcprior.mcmc.summer <- apply(mcmc,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.summer <- apply(mcmc,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.summer <- apply(mcmc,1,quantile,0.975, na.rm=TRUE)
fit.pcprior.mcmc.lin.summer <- apply(mcmc.lin,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.lin.summer <- apply(mcmc.lin,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.lin.summer <- apply(mcmc.lin,1,quantile,0.975, na.rm=TRUE)
partial.fit.out.summer <- X[,-6]%*%apply(samples.gamma[-6,], 1, mean) + 
  B[,-ind.summer]%*%apply(samples.beta[-ind.summer,], 1, mean)
partial.res.out.summer <- y - (partial.fit.out.summer)

# autumn
mcmc <- matrix(c(samples.gamma[7,]) %x% xg, 500, n.mcmc) + Bg%*%samples.beta.autumn
mcmc.lin <- matrix(c(samples.gamma[7,]) %x% xg, 500, n.mcmc) 
fit.pcprior.mcmc.autumn <- apply(mcmc,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.autumn <- apply(mcmc,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.autumn <- apply(mcmc,1,quantile,0.975, na.rm=TRUE)
fit.pcprior.mcmc.lin.autumn <- apply(mcmc.lin,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.lin.autumn <- apply(mcmc.lin,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.lin.autumn <- apply(mcmc.lin,1,quantile,0.975, na.rm=TRUE)
partial.fit.out.autumn <- X[,-7]%*%apply(samples.gamma[-7,], 1, mean) + 
  B[,-ind.autumn]%*%apply(samples.beta[-ind.autumn,], 1, mean)
partial.res.out.autumn <- y - (partial.fit.out.autumn)

# winter
mcmc <- matrix(c(samples.gamma[8,]) %x% xg, 500, n.mcmc) + Bg%*%samples.beta.winter
mcmc.lin <- matrix(c(samples.gamma[8,]) %x% xg, 500, n.mcmc) 
fit.pcprior.mcmc.winter <- apply(mcmc,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.winter <- apply(mcmc,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.winter <- apply(mcmc,1,quantile,0.975, na.rm=TRUE)
fit.pcprior.mcmc.lin.winter <- apply(mcmc.lin,1,mean, na.rm=TRUE)
q025.pcprior.mcmc.lin.winter <- apply(mcmc.lin,1,quantile,0.025,na.rm=TRUE)
q975.pcprior.mcmc.lin.winter <- apply(mcmc.lin,1,quantile,0.975, na.rm=TRUE)
partial.fit.out.winter <- X[,-8]%*%apply(samples.gamma[-8,], 1, mean) + 
  B[,-ind.winter]%*%apply(samples.beta[-ind.winter,], 1, mean)
partial.res.out.winter <- y - (partial.fit.out.winter)





# reproduce FIG 6 in the paper (it will be different as here we are using the fake data)

# plot smooth effect in spring
par(mfrow=c(1,1))
plot(xg-min(x),fit.pcprior.mcmc.spring, lty=1, lwd=2, type="l", ylim=c(0,6.5), 
     main="Spring", xlab="stream", ylab="f(stream)", cex.lab=1.5, cex.main=1.5)
polygon(c(rev(xg-min(x)), xg-min(x)), 
        c(rev(q975.pcprior.mcmc.spring), q025.pcprior.mcmc.spring), 
        col = 'grey80', border = NA)
lines(xg-min(x),fit.pcprior.mcmc.spring, lty=1, lwd=2, col=1)
points(x[dat$season=="spring"]-min(x), 
       partial.res.out.spring[dat$season=="spring"], col="grey28")

# plot smooth effect in summer
plot(xg-min(x),fit.pcprior.mcmc.summer, lty=1, lwd=2, type="l", ylim=c(0,6.5), 
     main="Summer", xlab="stream", ylab="f(stream)", cex.lab=1.5, cex.main=1.5)
polygon(c(rev(xg-min(x)), xg-min(x)), 
        c(rev(q975.pcprior.mcmc.summer), q025.pcprior.mcmc.summer), 
        col = 'grey80', border = NA)
lines(xg-min(x),fit.pcprior.mcmc.summer, lty=1, lwd=2, col=1)
points(x[dat$season=="summer"]-min(x), 
       partial.res.out.summer[dat$season=="summer"], col="grey28")

# plot smooth effect in autumn
plot(xg-min(x),fit.pcprior.mcmc.autumn, lty=1, lwd=2, type="l", ylim=c(0,6.5), 
     main="Autumn", xlab="stream", ylab="f(stream)", cex.lab=1.5, cex.main=1.5)
polygon(c(rev(xg-min(x)), xg-min(x)), 
        c(rev(q975.pcprior.mcmc.autumn), q025.pcprior.mcmc.autumn), 
        col = 'grey80', border = NA)
lines(xg-min(x),fit.pcprior.mcmc.autumn, lty=1, lwd=2, col=1)
points(x[dat$season=="autumn"]-min(x), 
       partial.res.out.autumn[dat$season=="autumn"], col="grey28")

# plot smooth effect in winter
plot(xg-min(x),fit.pcprior.mcmc.winter, lty=1, lwd=2, type="l", ylim=c(0,6.5), 
     main="Winter", xlab="stream", ylab="f(stream)", cex.lab=1.5, cex.main=1.5)
polygon(c(rev(xg-min(x)), xg-min(x)), 
        c(rev(q975.pcprior.mcmc.winter), q025.pcprior.mcmc.winter), 
        col = 'grey80', border = NA)
lines(xg-min(x),fit.pcprior.mcmc.winter, lty=1, lwd=2, col=1)
points(x[dat$season=="winter"]-min(x), 
       partial.res.out.winter[dat$season=="winter"], col="grey28")


