# this code reproduces Figures 1 to 4 in the paper (in color version)

source("utilities.R")
library(MASS)
library(spam)

## FIG 1: SCALING ISSUE
# marginal variances for different designs:
# 4 different B-spline design, changing K (num basis) and n (num covariate values)
# K={10,20}
# n={50,100}
# we use cubic B-splines

ndata <- 1000
ord <- 2

#K=50 
nseg <- 47  # number of segments between nodes
xl <- 0  
xr <- 1
eval.points <- seq(xl, xr, length.out=ndata)  
bdeg <- 3  # 1= tent basis; 3= cubic b-spline;  2=quadratic b-spline;   
dx <- (xr - xl) / nseg
nodes <- seq(xl - bdeg * dx, xr + bdeg * dx, by = dx)
B <- bspline(x=eval.points, xl=xl, xr=xr, ndx=nseg, bdeg=bdeg)
ncol(B) ; length(nodes)
D <- diff(diag(ncol(B)), diff=ord)
R <- crossprod(D,D)
# compute marginal variances under assumption that error precision (tau.e) is 1
Vstar <- B %*% ginv(R) %*% t(B)
plot(eval.points, sqrt(diag(Vstar)), ylim=c(0,90), 
     main ="K=50", xlab="x", ylab="standard deviation", type="l", cex.lab=1.5, cex.main=1.5)
# plot basis
for(i in 1:ncol(B)) lines(eval.points, 7*B[,i], col=i)

#K=100 
nseg <- 97  
xl <- 0  
xr <- 1
eval.points <- seq(xl, xr, length.out=ndata)  
bdeg <- 3    
dx <- (xr - xl) / nseg
nodes <- seq(xl - bdeg * dx, xr + bdeg * dx, by = dx)
B <- bspline(x=eval.points, xl=xl, xr=xr, ndx=nseg, bdeg=bdeg)
ncol(B) ; length(nodes)
D <- diff(diag(ncol(B)), diff=ord)
R <- crossprod(D,D)
Vstar <- B %*% ginv(R) %*% t(B)
plot(eval.points, sqrt(diag(Vstar)), ylim=c(0,90), 
     main ="K=100", xlab="x", ylab="standard deviation", type="l", cex.lab=1.5, cex.main=1.5)
for(i in 1:ncol(B)) lines(eval.points, 7*B[,i], col=i)




### FIG 2 
## evaluate the mapping for degrees of freedom,
# in the different designs, assuming tau.e=1

# K=10, n=50
map1 <- h.tau(x=seq(xl, xr, length.out=50), xl, xr, 
              nseg=7, bdeg=3, diff.ord=2, tau.e=1)
# K=10, n=100
map2 <- h.tau(x=seq(xl, xr, length.out=100), xl, xr, 
              nseg=7, bdeg=3, diff.ord=2, tau.e=1)
# K=20, n=50
map3 <- h.tau(x=seq(xl, xr, length.out=50), xl, xr, 
              nseg=17, bdeg=3, diff.ord=2, tau.e=1)
# K=20, n=100
map4 <- h.tau(x=seq(xl, xr, length.out=100), xl, xr, 
              nseg=17, bdeg=3, diff.ord=2, tau.e=1)

a1 <- -12
a2 <- 12
plot(splinefun(log(map1$tau.eval), map1$df.eval), ylim=c(0,20), xlim=c(a1,a2), 
     xlab=expression(log(tau[beta])), 
     main="",
     ylab="Degrees of freedom", col=1, cex.lab=1.5, cex.main=1.5)
plot(splinefun(log(map2$tau.eval), map2$df.eval), ylim=c(0,20), xlim=c(a1,a2), col=2, add=TRUE)
plot(splinefun(log(map3$tau.eval), map3$df.eval), ylim=c(0,20), xlim=c(a1,a2), col=3, add=TRUE)
plot(splinefun(log(map4$tau.eval), map4$df.eval), ylim=c(0,20), xlim=c(a1,a2), col=4, add=TRUE)
legend("topright", legend=c("K = 10; n = 50", "K = 10; n = 100", "K = 20; n = 50", "K = 20; n = 100"),
       fill=c(1,2,3,4))
abline(h=2, lty=3, lwd=3)


## perspective plot: degrees of freedom vs tau.beta and tau.e
# h.tau0() below is the same as h.tau(), but evaluates the mapping at 50 locations instead of 1000
h.tau0 <- function(x, xl=min(x), xr=max(x), nseg, bdeg, diff.ord, tau.e, eps.diag=0){
  B <- bspline(x=x, xl=xl, xr=xr, ndx=nseg, bdeg=bdeg)
  D <- diff(diag(ncol(B)), diff=diff.ord)
  R <- crossprod(D,D)
  BtB <- crossprod(B,B) + diag(rep(eps.diag), ncol(B))
  d0 <- eigen(solve(BtB)%*%R)$val
  d <- d0*(1/tau.e)  
  tau.eval <- c(exp(seq(-20,20,length.out=50)))
  edf.tau <- rep(NA, length(tau.eval))
  for(i in 1:length(tau.eval)) 
    edf.tau[i] <- sum(1/(1+d*tau.eval[i]))
  f.tau <- splinefun(tau.eval, edf.tau)
  finv.edf <- splinefun(edf.tau, tau.eval)
  list(df=f.tau, tau=finv.edf, rd=diff.ord, K=ncol(B), df.eval=edf.tau, tau.eval=tau.eval, B=B, R=R, v0=d0)
}

tau.epsilon <- exp(seq(-10,10, length.out=50))
grid.z <- matrix(, nrow=length(tau.epsilon), ncol=50)
for(i in 1:length(tau.epsilon)){
   map <- h.tau0(x=seq(xl, xr, length.out=50), xl, xr, 
                             nseg=17, bdeg=3, diff.ord=2, tau.e=tau.epsilon[i], eps.diag=.001)
   grid.z[i,] <- map$df.eval 
}
grid.tau.b <- log(map$tau.eval)

persp(grid.tau.b, log(tau.epsilon), t(grid.z), ticktype="detailed", col="white", nticks=4, 
      theta=20, phi=40, ltheta = 60, lphi=60, shade = .2, expand = 0.4,
      xlab=expression(log(tau[beta])), ylab=expression(log(tau[epsilon])), zlab="Degrees of freedom", 
      zlim=c(0,22), cex.main=1.5, cex.lab=1.5, 
      main="")



### FIG 3
# plot the induced PC pior for deegres of freedom with alpha=0.01 and U=5, 
# consider the 4 mappings computed above (map1, map2, map3, map4)

alpha<-.01
U <- 5
u.map <-  1/sqrt(map1$tau(U))
theta.map1 <- -log(alpha)/u.map
pi.df_pcprior.map1 <- PCprior.implied.df(tau=map1$tau.eval, df=map1$df.eval, theta=theta.map1)
u.map <-  1/sqrt(map2$tau(U))
theta.map2 <- -log(alpha)/u.map
pi.df_pcprior.map2 <- PCprior.implied.df(tau=map2$tau.eval, df=map2$df.eval, theta=theta.map2)
u.map <-  1/sqrt(map3$tau(U))
theta.map3 <- -log(alpha)/u.map
pi.df_pcprior.map3 <- PCprior.implied.df(tau=map3$tau.eval, df=map3$df.eval, theta=theta.map3)
u.map <-  1/sqrt(map4$tau(U))
theta.map4 <- -log(alpha)/u.map
pi.df_pcprior.map4 <- PCprior.implied.df(tau=map4$tau.eval, df=map4$df.eval, theta=theta.map4)


# plot PC prior for all mappings on log tau scale (draft-pcprior-logtau.ps)
plot(seq(-20,30,length.out=10000), PCprior.logtau(seq(-20,30,length.out=10000), theta.map1), 
     type="l", col=1, ylab="density", xlab=expression(log(tau[beta])), xlim=c(-5,20),
     main="",
     cex.lab=1.5, cex.main=1.5)
lines(seq(-20,30,length.out=10000), PCprior.logtau(seq(-20,30,length.out=10000), theta.map2), col=2)
lines(seq(-20,30,length.out=10000), PCprior.logtau(seq(-20,30,length.out=10000), theta.map3), col=3)
lines(seq(-20,30,length.out=10000), PCprior.logtau(seq(-20,30,length.out=10000), theta.map4), col=4)
legend("topright", legend=c("K = 10; n = 50", "K = 10; n = 100", "K = 20; n = 50", "K = 20; n = 100"),
       fill=c(1,2,3,4))


# plot PC prior for all mapppings on df scale (draft-pcprior-df.ps)
plot(pi.df_pcprior.map1$x, pi.df_pcprior.map1$y, type="l", lwd=1, xlim=c(1.99,6), ylim=c(0,1.5), xaxt="n",
      ylab="density", xlab=expression(d), 
     main="", 
      cex.lab=1.5, cex.main=1.5)
lines(pi.df_pcprior.map2$x, pi.df_pcprior.map2$y, lwd=1, col=2)
lines(pi.df_pcprior.map3$x, pi.df_pcprior.map3$y, lwd=1, col=3)
lines(pi.df_pcprior.map4$x, pi.df_pcprior.map4$y, lwd=1, col=4)
legend("topright", legend=c("K = 10; n = 50", "K = 10; n = 100", "K = 20; n = 50", "K = 20; n = 100"),
       fill=c(1,2,3,4))
abline(v=2, lty=3, lwd=3)
axis(side=1, at=c(2,3,4,6))
axis(side=1, at=5, lwd=2, labels="U = 5", col=2)



### FIG 4
# plot implied degrees of freedom for a Gamma(a,b) 

a <-  1
b <-  .0005
pi.df_gamma.map1 <- Gamma.implied.df(tau=map1$tau.eval, df=map1$df.eval, a=a, b=b)
pi.df_gamma.map2 <- Gamma.implied.df(tau=map2$tau.eval, df=map2$df.eval, a=a, b=b)
pi.df_gamma.map3 <- Gamma.implied.df(tau=map3$tau.eval, df=map3$df.eval, a=a, b=b)
pi.df_gamma.map4 <- Gamma.implied.df(tau=map4$tau.eval, df=map4$df.eval, a=a, b=b)
plot(pi.df_gamma.map1$x, pi.df_gamma.map1$y, xlim=c(1.99,4),type="l", 
     xlab=expression(d), ylab="density", 
     main="",
     cex.lab=1.5, cex.main=1.5)
lines(pi.df_gamma.map2$x, pi.df_gamma.map2$y, type="l", col=2, 
      xlab="df", ylab="pi(df)", main="")
lines(pi.df_gamma.map3$x, pi.df_gamma.map3$y, type="l", col=3, 
      xlab="df", ylab="pi(df)", main="")
lines(pi.df_gamma.map4$x, pi.df_gamma.map4$y, type="l", col=4, 
      xlab="df", ylab="pi(df)", main="")
legend("topright", legend=c("K = 10; n = 50", "K = 10; n = 100", "K = 20; n = 50", "K = 20; n = 100"),
       fill=c(1,2,3,4), border = "black", bty="n")


a <- .001
b <- .001
pi.df_gamma.map1 <- Gamma.implied.df(tau=map1$tau.eval, df=map1$df.eval, a=a, b=b)
pi.df_gamma.map2 <- Gamma.implied.df(tau=map2$tau.eval, df=map2$df.eval, a=a, b=b)
pi.df_gamma.map3 <- Gamma.implied.df(tau=map3$tau.eval, df=map3$df.eval, a=a, b=b)
pi.df_gamma.map4 <- Gamma.implied.df(tau=map4$tau.eval, df=map4$df.eval, a=a, b=b)
plot(pi.df_gamma.map1$x, pi.df_gamma.map1$y, xlim=c(1.99,20), ylim=c(0,.0099), type="l", 
     xlab=expression(d), ylab="density", 
     main="",
     cex.lab=1.5, cex.main=1.5)
lines(pi.df_gamma.map2$x, pi.df_gamma.map2$y, type="l", col=2, 
       xlab="df", ylab="pi(df)", main="")
lines(pi.df_gamma.map3$x, pi.df_gamma.map3$y, type="l", col=3, 
       xlab="df", ylab="pi(df)", main="")
lines(pi.df_gamma.map4$x, pi.df_gamma.map4$y, type="l", col=4, 
       xlab="df", ylab="pi(df)", main="")
legend(x=12, y=0.01, legend=c("K = 10; n = 50", "K = 10; n = 100", "K = 20; n = 50", "K = 20; n = 100"),
       fill=c(1,2,3,4), border = "black", bty="n")



### FIG 5 in the supplemental material
# show the PC prior for degrees of freedom in action with simulated data 
# we show the code for only scenario 1, for scenario 2 and 3 is the same..

ndata <- 50
tau.e <- .01
xl <- 0  
xr <- 5
eval.points <- seq(xl, xr, length.out=ndata)  

# generate data from scenario 1: polynomial degree 3
y.true <- 5+eval.points + .2*(eval.points)^2 + .3*(eval.points)^3   
set.seed(1) 
y <- y.true + rnorm(length(eval.points), 0, 1/sqrt(tau.e)) 

nseg <- 27
bdeg <- 3
ord <- 2
map <- h.tau(x=eval.points, xl=min(eval.points), xr=max(eval.points), 
             nseg=nseg, bdeg=bdeg, diff.ord=ord, tau.e=tau.e)
B <- map$B
R <- map$R

# set the PC prior parameters: alpha and the upper bounds for d (i.e. U)
# in the Figure in the paper we take different values for U, i.e. {2,4,6}, 
# in this example we take U=4 
# (NOTE: if you want to assume U=2, in order to avoid numerical issue fix U=2 + small eps..)
alpha <- .01
U <- 4


#### fit the model with MCMC algorithm 1 in the Appendix 
burnin <- 1:1000
ini <- list(beta=rep(0,ncol(B)),
                tau.e=1,
                tau.b=10)

result <- mcmc.pcprior(y=y, 
                   P.spline=list(x=eval.points, xl=min(eval.points), xr=max(eval.points), 
                                 nseg=nseg, bdeg=bdeg, ord=ord), 
                   pcprior=list(U=U, alpha=alpha), 
                   init=list(beta=ini$beta, 
                             tau.e=ini$tau.e, 
                             tau.b=ini$tau.b), 
                   mcmc=list(n.iter=5000, tun=1.3), 
                   verbose=TRUE)

hat.beta <- apply(result$samples.beta[,-burnin], 1, mean, na.rm=T)

plot(eval.points, y, xlab="x", ylab="y")
lines(eval.points, y.true, lwd=2, lty=2)
lines(eval.points, B%*%hat.beta, lwd=2, col=3)




