
library(splines)

# in Eilers 1996 viene useto il package splines e la seguente funzione
bspline <- function(x, xl, xr, ndx, bdeg) {
  dx <- (xr - xl) / ndx
  knots <- seq(xl - bdeg * dx, xr + bdeg * dx, by = dx)
  B <- spline.des(knots, x, bdeg + 1, 0 * x, outer.ok = TRUE)$design
  B
}


# below find some functions to compute the prior for degrees of freedom, pi(df)
# implied by the prior for tau, pi(tau) 
# possible choiches of pi(tau):
# the Gumbel2(a=1/2, b=theta), i.e. the PC prior for tau 
# the conjugate Gamma(a, b)
# in input we need the mapping df=h(tau), i.e. a vector of tau's and a vector of h(tau) 
# return the density pi(df) implied by the assumed prior for tau

# Gumbel type 2: the PC-prior for tau
PCprior.tau <- function(tau, theta){
   .5*theta*tau^(-1.5) * exp(-theta*tau^(-.5))
}

# Gumbel type 2: the PC-prior for tau (in the log scale)
PCprior.logtau <- function(logtau, theta){
  .5*theta*exp(logtau)^(-.5) * exp(-theta*exp(logtau)^(-.5))
}

# pi(df) calculated as the 'df' implied by a PC prior for the precision tau
PCprior.implied.df <- function(tau, df, theta){
  der.inv <- diff(tau)/diff(df)           # der.inv calculated at interpoints between tau values
  f.inv <- tau[-length(tau)] + diff(tau)  # f.inv calculated at interpoints between tau values
  list(x=df[-length(df)] + diff(df), y=PCprior.tau(f.inv, theta) * abs(der.inv))
}


# pi(df) calculated as the 'df' implied by a Gamma(a,b) prior for the precision tau
Gamma.implied.df <- function(tau, df, a, b){
  der.inv <- diff(tau)/diff(df)           
  f.inv <- tau[-length(tau)] + diff(tau)  
  list(x=df[-length(df)] + diff(df), y=dgamma(f.inv, shape=a, rate=b) * abs(der.inv))
}


### mapping df=h(tau), given the b-spline design information and the random walk order
# nseg: number of segments between equispaced knots
# bdeg: degree of the b-splines
# diff.ord: order of the RW model on the spline coefs, i.e. rank deficiency of the random walk
# tau.e: the scale of the data
h.tau <- function(x, xl=min(x), xr=max(x), nseg, bdeg, diff.ord, tau.e, eps.diag=0){
  # build b-splines
  B <- bspline(x=x, xl=xl, xr=xr, ndx=nseg, bdeg=bdeg)
  
  # build structure matrix
  D <- diff(diag(ncol(B)), diff=diff.ord)
  R <- crossprod(D,D)
  
  # compute mapping df=h(tau), assuming prec.y is known
  BtB <- crossprod(B,B) + diag(rep(eps.diag), ncol(B))
  d0 <- eigen(solve(BtB)%*%R)$val
  d <- d0*(1/tau.e)  # scaling according to tau.e
  
  # map tau --> edf : it is convenient to build the grid of tau valuesby taking the exp of log(tau) defined inside interval (-20,20)
  tau.eval <- c(exp(seq(-20,20,length.out=1000)))
  edf.tau <- rep(NA, length(tau.eval))
  for(i in 1:length(tau.eval)) 
  edf.tau[i] <- sum(1/(1+d*tau.eval[i]))
  f.tau <- splinefun(tau.eval, edf.tau)
  finv.edf <- splinefun(edf.tau, tau.eval)
  list(df=f.tau, tau=finv.edf, rd=diff.ord, K=ncol(B), df.eval=edf.tau, tau.eval=tau.eval, B=B, R=R, v0=d0)
}


# mapping df=h(tau) given the 'v0'=eigenvalues of R%*%(B^T B)^-1 and tau.e=1/noise variance  
h.tau.given.scale <- function(v0, tau.e){
  # compute mapping df=h(tau), assuming prec.y is known
  #v0 <- eigen(solve(crossprod(B,B))%*%R)$val
  v <- v0*(1/tau.e)  # scaling according to tau.e
  
  # find the mapping on the log scale
  tau.eval <- c(exp(seq(-20,20,length.out=1000)))
  edf.tau <- rep(NA, length(tau.eval))
  for(i in 1:length(tau.eval)) 
    edf.tau[i] <- sum(1/(1+v*tau.eval[i]))
  f.tau <- splinefun(tau.eval, edf.tau)
  finv.edf <- splinefun(edf.tau, tau.eval)
  list(df=f.tau, tau=finv.edf, df.eval=edf.tau, tau.eval=tau.eval)
}






#### mcmc

# function by Haavard for sampling pi(f) \propto 1 + 1/f   for f in [1/F, F]
ranf = function(n=1, F=2)
{
  if (F <= 1.0) {
    return (rep(1.0, n))
  } else {
    len = F - 1.0 / F;
    
    nx = rbinom(1, size=n,  prob= len / (len + 2.0 * log(F)))
    ny = n - nx
    
    if (nx > 0L) {
      x = 1/F + runif(nx, max = len)
    } else {
      x = NULL
    }
    if (ny > 0L) {
      y = F^runif(ny, min=-1, max=1)
    } else {
      y = NULL
    }
    
    return (c(x, y))
  }
}


sample.gmrf <- function(Q, b){
  L <- chol.spam(as.spam(Q)) 
  z <- rnorm(ncol(Q)) 
  mean.beta <- backsolve(L, forwardsolve(L, b)) 
  beta <- mean.beta + backsolve(L, z)
  list(beta=beta, z=z, L=L)
}

# this is equal to sample.gmrf, but it does not use spam (it is good enough for sampling the fixed effects though)
sample.fixed <- function(Q, b){
  L <- chol(Q)
  z <- rnorm(ncol(Q)) 
  gamma <- backsolve(L, backsolve(L, b, transpose=TRUE) + z)  
  list(gamma=gamma, z=z, L=L)
}

## functions for sampling GMRF under linear constr; works for rankdef=2 (i.e. 2nd order RW)
sample.gmrf.ULC <- function(Q, b, A){
  e <- matrix(c(0,0),2,1)    
  L <- chol.spam(as.spam(Q)) 
  z <- rnorm(ncol(Q)) 
  mean.beta <- backsolve(L, forwardsolve(L, b)) 
  beta <- mean.beta + backsolve(L, z)    #backsolve(L, forwardsolve(L, b)+z)
  V <- cbind(backsolve(L, forwardsolve(L, A[1,])), backsolve(L, forwardsolve(L, A[2,])))
  W <- A%*%V
  UU <- solve(W) %*% t(V)
  cc <- A%*%beta - e
  beta <- c(beta - t(UU)%*%cc)
  list(beta=beta, mean.beta=mean.beta, z=z, L=L)
}


## this function is involved in evaluating the acceptance ratio for block updating (tau.e,tau.b,beta) 
## in ALGORITHM 1
# computes log joint posterior distribution of tau.b, tau.e, pi(tau.e,tau.b|y)
# this version is efficient because it uses the cholesky decomoposition of the posterior precision Q
# to compute the log full conditional for beta:
# 'L' is the cholesky of Q, computed using spam package
# 'q' is defined in formula (2.26) in Haavard's book
#  q=sum(z^2), where z are the samples from the standard normal involved in sampling beta~N(Q^-1 %*% b, Q^-1)
log.joint.taus.chol <- function(y, mean.y, B, BtB, tau.e, beta, tau.b, R, theta.e, rankdef=2, L, q){
  n <- length(y)
  K <- length(beta)
  
  # compute log likelihood
  log.lik <- -n*0.5*log(2*pi) + 0.5*n*log(tau.e)- 0.5*tau.e*sum((y-mean.y)^2)
  
  # compute log prior density for beta|tau.b, i.e. log pi(beta|tau.b)
  log.det.precmat <- 0.5*sum(log(eigen(tau.b*R, only.values = TRUE)$values[1:(ncol(R)-rankdef)]))  
  log.beta.given.tau.b <- -K*0.5*log(2*pi) + log.det.precmat - 0.5*tau.b*(beta%*%R%*%beta) 
  
  # compute log pi(tau.b|tau.e)
  log.tau.b <- log(PCprior.tau(tau.b, theta.e))
  
  # compute log pi(tau.e) = 1/tau.e
  log.tau.e <- -log(tau.e) 
  
  # compute log full conditional for beta, i.e. log pi(beta|tau.b,y)
  log.det.precmat <-  sum(log(diag(L)))
  log.fullcond.beta <-  -K*0.5*log(2*pi) + log.det.precmat - 0.5*q
  
  log.lik + log.beta.given.tau.b + log.tau.b + log.tau.e - log.fullcond.beta
}


## compute log full conditional for beta under linear constr, i.e. log pi(beta|tau.b,y)
log.fullcond.beta.ULC <- function(beta, mean.beta, Q, L, A){
  v <- (beta - mean.beta)
  q <- crossprod(v, Q%*%v)
  log.det.precmat <-  sum(log(diag(L)))
  log.x <-  log.det.precmat - 0.5*q         # log(pi(x|...))
  log.pi.Ax.given.x <-  -sum(log(diag(chol(A%*%t(A)))))  # log(pi(Ax|x,...))
  mean.Ax <- A%*%mean.beta
  var.Ax <- A%*%solve(Q)%*%t(A)
  log.det.var.cov <- sum(log(diag(chol(var.Ax))))
  v2 <- (A%*%beta) - mean.Ax
  log.pi.Ax <- - log.det.var.cov - 0.5 * crossprod(v2, var.Ax%*%v2) # log(pi(Ax|x))
  log.x + log.pi.Ax.given.x - log.pi.Ax 
}



## this function is involved in evaluating the acceptance ratio for block updating (tau.e,gamma) 
## in ALGORITHM 2
# y: data
# tau.e: the scale of the errors, CANDIDATE value
# mu.tilde: the linear predictor - the fixed effect term under examination
# X: fixed effects design matrix
# gamma: vector of fixed effects, CANDIDATE value
# tau.g: prior precision of fixed effects, fixed at some small values (no prior on it)
# L: cholesky factor
# z: random uniform number involved in sampling the candidate fixed effects
marginal.prec.errors <- function(y, tau.e, mu.tilde, X, gamma, tau.g, L, z){
  n <- length(y)
  res <- y - c(X%*%gamma + mu.tilde)
  log.lik <-  0.5*n*log(tau.e)- 0.5*tau.e*sum(res^2) 
  log.gamma <- -0.5*tau.g*crossprod(gamma, gamma)  
  log.tau.e <- -log(tau.e) # jeffreys prior
  log.det.precmat <-  sum(log(diag(L)))
  log.fullcond.gamma <- + log.det.precmat - 0.5*sum(z^2) 
  log.marginal.tau.e <- log.lik + log.gamma + log.tau.e - log.fullcond.gamma
  log.marginal.tau.e
}


## function 'mcmc.pcprior()' reproduces ALGORITHM 1
## MCMC with joint update (tau.e, tau.b, beta) using a joint (PC)prior for (tau.b,tau.e)
## one covariate p-spline model, with an IGMRF on beta (i.e. no intercept and thus no need for constraints)
# 'init' is a list with initial values for vector 'beta', the precisions 'tau.e' and 'tau.b' 
# 'pcprior' is a list with pcprior user scaling parameters 'U' and 'alpha' 
# 'mcmc' is a list with 
#    'n.iter'
#    'tun' is a tuning parameter for the proposal for the precisions (4.13 in Haavard's book) (NOTE, 'ranf()' is Haavard's function)
# 'P.spline' is a list with 
#     'x', is a vector with covariate values  
#     'xl' is the min covariate range value 
#     'xr, is the max covariate range value 
#     'nseg' is the number of equally-length segment over the covariate domain 
#     'bdeg' is the degree of the b-spline basis functions
#     'ord' is the order of the random walk model assumed for the spline coefs 
mcmc.pcprior <- function(y, P.spline, pcprior, init, mcmc, eps.diag=0, verbose=TRUE){
  
  x <- P.spline$x
  xl <- P.spline$xl
  xr <- P.spline$xr
  nseg <- P.spline$nseg
  bdeg <- P.spline$bdeg
  ord <- P.spline$ord
  tau.e <- init$tau.e 
  tau.b <- init$tau.b 
  beta <- init$beta
  
  # scale the pcprior according to the preset user-scaling U and alpha
  map <- h.tau(x=x, nseg=nseg, bdeg=bdeg, diff.ord=ord, tau.e=tau.e, eps.diag=eps.diag)
  alpha <- pcprior$alpha
  U <- pcprior$U
  theta.e <- -log(alpha)*sqrt(map$tau(U))
  
  # b-spline design matrix
  B <- bspline(x=x, xl=xl, xr=xr, ndx=nseg, bdeg=bdeg)
  K <- ncol(B)
  BtB <- crossprod(B,B)
  Bty <- crossprod(B,y)
  
  # build precision structure
  D <- diff(diag(ncol(B)), diff=ord)
  R <- crossprod(D,D)
  
  # compute log.joint.taus.current for the initial values, see Haavard's book (2.27)
  b <- tau.e*Bty
  Q <- tau.e*BtB + tau.b*R
  L <- chol.spam(as.spam(Q))
  mean.y <- c(B%*%beta)
  mu.beta <- backsolve(L, forwardsolve(L, b))
  v <- beta - mu.beta  
  w <- Q %*% v
  q <- crossprod(v, w)
  log.joint.taus.current <- log.joint.taus.chol(y, mean.y, B, BtB, tau.e, beta, tau.b, R, theta.e, rankdef=ord, L, q)
  
  # build empty mcmm boxes
  samples.beta <- matrix(nrow=K, ncol=1)
  samples.tau.e <- matrix(, nrow=1, ncol=1)
  samples.tau.b <- matrix(, nrow=1, ncol=1)
  acceptance <- rep(0, mcmc$n.iter)
  
  for (i in 1:mcmc$n.iter) {
    
    ### the whole mcmc is an M-H to accept/reject (tau.e, tau.b, beta) jointly 
    
    # first, sample a candidate for tau.b and tau.e from proposal 4.13 in Haavard's book
    tau.e.cand <- tau.e * ranf(1,mcmc$tun)  
    tau.b.cand <- tau.b * ranf(1,mcmc$tun)
    # second, sample beta|tau.b.cand (IMP! the proposal for beta is its own full conditional!)
    b <- tau.e.cand*Bty
    Q <- tau.e.cand*BtB + tau.b.cand*R
    L <- chol.spam(as.spam(Q))
    z <- rnorm(K) 
    beta.cand <- backsolve(L, forwardsolve(L, b)+z)
    q <- sum(z^2)
    mean.y <- c(B%*%beta.cand)
    # before accepting/rejecting we need to scale the PC-prior parameter 'theta' according to tau.e.cand
    finvU <- (h.tau.given.scale(map$v0, tau.e.cand)$tau(U))    
    theta.e.cand <- -log(alpha)*sqrt(finvU)
    
    # accepting/rejecting jointly (tau.e.cand, tau.b.cand, beta.cand)
    log.joint.taus.new <- log.joint.taus.chol(y, mean.y, B, BtB, tau.e.cand, beta.cand, tau.b.cand, R, theta.e.cand, rankdef=ord, L, q)
    if (runif(1) < exp(min(0, log.joint.taus.new - log.joint.taus.current ))){
      log.joint.taus.current <- log.joint.taus.new
      tau.e <- tau.e.cand
      tau.b <- tau.b.cand
      beta <- beta.cand
      acceptance[i] <- 1
    } 
    
    # tuning the proposal? no it should be fine with 'tun' around 1.5
    k <- i/100;  
    if(verbose & (ceiling(k) == floor(k))) {
      accept.rate <- mean(acceptance[(i-100+1):i])*100
      print(paste("n.iter ", i,"  ; tun=", mcmc$tun,"  ; accept rate=",accept.rate, "  ; min tau.b=",min(samples.tau.b, na.rm=T),
                  "  ; max tau.b=", max(samples.tau.b, na.rm=T)))
      print(paste("n.iter ", i,"  ; tun=", mcmc$tun,"  ; accept rate=",accept.rate, "  ; min tau.e=",min(samples.tau.e, na.rm=T),
                  "  ; max tau.e=", max(samples.tau.e, na.rm=T)))
    }
    
    # collect samples
    samples.tau.e <- cbind(samples.tau.e, tau.e)
    samples.tau.b <- cbind(samples.tau.b, tau.b)
    samples.beta <- cbind(samples.beta, beta)
  }
  
  list(y=y, P.spline=P.spline, pcprior=pcprior, init=init, mcmc=mcmc, last=c(beta, tau.b, tau.e), 
       samples.beta=samples.beta, samples.tau.b=samples.tau.b, samples.tau.e=samples.tau.e)
}



### function below is used to fit the additive p-spline model in section 7
# see file 'code application with fake data.R'
# compute marginal precision for beta (the GMRF for all the 4 seasons)
marginal.prec.splinecoef.allthe4.ULC <- 
  function(y, tau.e, mu.tilde, B, beta, mean.beta, Q, L, A, tau.b, R0, ord, theta.e){ 
    n <- length(y)
    res <- y - c(mu.tilde+B%*%beta)
    log.lik <-  0.5*n*log(tau.e)- 0.5*tau.e*sum(res^2) 
    det.tmp <- sum(log(eigen(tau.b[1]*R0, only.values = TRUE)$values[1:(ncol(R0)-ord)])) 
    det.tmp <- det.tmp + sum(log(eigen(tau.b[2]*R0, only.values = TRUE)$values[1:(ncol(R0)-ord)])) 
    det.tmp <- det.tmp + sum(log(eigen(tau.b[3]*R0, only.values = TRUE)$values[1:(ncol(R0)-ord)])) 
    det.tmp <- det.tmp + sum(log(eigen(tau.b[4]*R0, only.values = TRUE)$values[1:(ncol(R0)-ord)])) 
    # compute log.tau.b by adding log.tau.b1+...+log.tau.b4
    log.det.precmat <- 0.5*det.tmp
    log.beta.given.tau.b <- log.det.precmat - 0.5*(beta%*%(diag(tau.b)%x%R0)%*%beta) 
    log.tau.b1 <- log(PCprior.tau(tau.b[1], theta.e[1]))
    log.tau.b2 <- log(PCprior.tau(tau.b[2], theta.e[2]))
    log.tau.b3 <- log(PCprior.tau(tau.b[3], theta.e[3]))
    log.tau.b4 <- log(PCprior.tau(tau.b[4], theta.e[4]))
    log.tau.b <- log.tau.b1 + log.tau.b2 + log.tau.b3 + log.tau.b4
    log.fullcond.beta <- log.fullcond.beta.ULC(beta, mean.beta, Q, L, A) 
    log.marginal.tau.b <- log.lik + log.beta.given.tau.b + log.tau.b - log.fullcond.beta
    log.marginal.tau.b
  }













